
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "mem_map.h"

unsigned map(int memfd, unsigned long dev_base, mem_map_t *m) {

    // Map one page of memory into user space such that the device is in that page, but it may not
    // be at the start of the page.

    m->base = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, dev_base & ~MAP_MASK);
        if (m->base == (void *) -1) {
        printf("Can't map the memory to user space.\n");
        return 1;
    }
    printf("Memory mapped at address %p.\n", m->base);

    // get the address of the device in user space which will be an offset from the base
    // that was mapped as memory is mapped at the start of a page

    m->dev_base = ((UINTPTR)m->base + (dev_base & MAP_MASK));

    return 0;
}

/*
 * Framebuffer memory mapping (assumes each frame start aligned to 4096 bytes
 */
unsigned maps(int memfd, unsigned long base, mem_map_t *m, unsigned long map_size) {

	// Map one page of memory into user space such that the device is in that page, but it may not
	// be at the start of the page.

	if ((base % MAP_SIZE) != 0) {
		printf("Start of memory block must be aligned to 4096 bytes\r\n");
		return 1;
	}

	m->base = mmap(0, map_size, PROT_READ | PROT_WRITE, MAP_SHARED, memfd,
			base);
	if (m->base == (void *) -1) {
		printf("Can't map the memory to user space.\n");
		return 1;
	}
	printf("Memory mapped at address %p.\n", m->base);

	// get the address of the device in user space which will be an offset from the base
	// that was mapped as memory is mapped at the start of a page

	//m->dev_base = m->base + (dev_base & MAP_MASK);
	m->dev_base = (UINTPTR) m->base; //(void*)(char*)(m->base);

	return 0;
}
