
#include <stdint.h>
#include <stdio.h>
#include <iostream>
#include <cstdio>
#include <stdlib.h>
#include <xil_io.h>
#include <unistd.h>
#include "xparameters.h"
#include "xmutex.h"

#include "optflow.h"
#include "sds_utils.h"

#define HEIGHT 1080
#define WIDTH 1920
#define FILEINAME "route85_1920x1080.yuv"
#define ONAME "OptFlow"
	
pix_t fileBuf[WIDTH];

void read_yuv_frame(pix_t *buf, int s, int w, int h, FILE *f);
void write_yuv_file(yuv_t *buf, int s, int w, int h, char *name);
void convert_Y8toCY16(pix_t *Ybuf, yuv_t *CYbuf, int pixct);
void convert_CY16toY8(yuv_t *CYbuf, pix_t *Ybuf, int pixct);

#include "mem_map.h"

#define MUTEX_DEVICE_ID		XPAR_MUTEX_0_IF_0_DEVICE_ID

/*
 * The Mutex number on which the example is run
 */
#define MUTEX_NUM 		0

/*
 * Mutex instance
 */
XMutex Mutex;

void start_Microblaze() {
    // file handles
    int valuefd, exportfd, directionfd;

    // The GPIO has to be exported to be able to see it
    // in sysfs

    exportfd = open("/sys/class/gpio/export", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to export\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);

    // Update the direction of the GPIO to be an output

    directionfd = open("/sys/class/gpio/gpio960/direction", O_RDWR);
    if (directionfd < 0)
    {
        printf("Cannot open GPIO direction it\n");
        exit(1);
    }

    write(directionfd, "out", 4);
    close(directionfd);

    // Get the GPIO value ready to be toggled

    valuefd = open("/sys/class/gpio/gpio960/value", O_RDWR);
    if (valuefd < 0)
    {
        printf("Cannot open GPIO value\n");
        exit(1);
    }

    // reset Microblaze
    write(valuefd,"0", 2);

    printf("Reset for 1 sec. ...\n");
    sleep(1);
    printf("Reset for 1 sec. done\n");

    // release MB reset
    write(valuefd,"1", 2);

    close(valuefd);

    //release exported gpio pin
    exportfd = open("/sys/class/gpio/unexport", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to unexport\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);
    return;
}

// monochrome yuv422 file is read using width w * height h
// only Y data is read, and returned in unsigned char buffer
// stride s is >= width w
// use buffer pix_t fileBuf[WIDTH];
// skip over U and V planes

void read_yuv_frame(pix_t *buf, int s, int w, int h, FILE *f)
{
    int i,j;
    
    // store Y
    for (j=0;j<h;j++) {
        fread(fileBuf, w, 1, f);
        for (i=0;i<w;i++) {
            *(buf+j*s+i) = fileBuf[i];
        }
    }

    // skip over U
    for (j=0;j<h;j++)
        fread(fileBuf, w/2, 1, f);

    // skip over V
    for (j=0;j<h;j++)
        fread(fileBuf, w/2, 1, f);

}

// monochrome yuv422 file is written from Y data using width w * height h
// w * h is appended to name, recognized by YUV Player Deluxe and other players
// stride s is >= width w
// use buffer unsigned char fileBuf[WIDTH];
void write_yuv_file(yuv_t *buf, int s, int w, int h, char *name)
{
    int i,j;
    char fname[32];
    FILE *f;
	yuv_t yuvval;
    
    sprintf(fname,"%s_%dx%d.yuv",name,w,h);
    f = fopen(fname,"wb");
    if (f==NULL) {
        printf("failed to open file %s\n", fname);
        return;
    }
    else {
        printf("wrote file %s\n", fname);
    }

    // Y
    for (j=0;j<h;j++) {
        for (i=0;i<w;i++) {
            yuvval = *(buf+j*s+i);
            fileBuf[i] = (pix_t) (yuvval & 0x00FF);
        }
        fwrite(fileBuf, w, 1, f);
    }

    // U
    for (j=0;j<h;j++) {
        for (i=0;i<w/2;i++) {
            yuvval = *(buf+j*s+(i*2));
            fileBuf[i] = (pix_t) ((yuvval>>8) & 0x00FF);
        }
        fwrite(fileBuf, w/2, 1, f);
    }

    // V
    for (j=0;j<h;j++) {
        for (i=0;i<w/2;i++) {
            yuvval = *(buf+j*s+(i*2+1));
            fileBuf[i] = (pix_t) ((yuvval>>8) & 0x00FF);
        }
        fwrite(fileBuf, w/2, 1, f);
    }

    fclose(f);
}

void convert_Y8toCY16(pix_t *Ybuf, yuv_t *CYbuf, int pixct)
{
	int i;
	yuv_t cypix;

	for (i=0; i<pixct; ++i) {
		cypix = 0x8000 | Ybuf[i];
		CYbuf[i] = cypix;
	}
}

int main()
{

	/*
	 * Init mutex and start microblaze
	 */
	XMutex_Config *ConfigPtr;
	XStatus Status;
	u32 TimeoutCount = 0;
	u32 val;

    u32 iiii=0;
    u32 value;

	int memfd;
	memfd = open("/dev/mem", O_RDWR | O_SYNC);
	if (memfd == -1) {
		printf("Can't open /dev/mem.\n");
		exit(0);
	}
	printf("/dev/mem opened.\n");

	/*
	 * Lookup configuration data in the device configuration table.
	 * Use this configuration info down below when initializing this
	 * driver instance.
	 */

	ConfigPtr = XMutex_LookupConfig(MUTEX_DEVICE_ID);
	if (ConfigPtr == (XMutex_Config *) NULL) {
		return XST_FAILURE;
	}

	/*
	 * Map Mutex memory to userspace
	 */
	mem_map_t mutex_map;

	if (map(memfd, (off_t) ConfigPtr->BaseAddress, &mutex_map)) {
		printf("Mutex mapping to userspace failed\r\n");
		exit(1);
	}

	ConfigPtr->BaseAddress = mutex_map.dev_base;

	/*
	 * Perform the rest of the initialization.
	 */
	Status = XMutex_CfgInitialize(&Mutex, ConfigPtr, ConfigPtr->BaseAddress);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XMutex_SetUser(&Mutex, MUTEX_NUM, 0);
	XMutex_Lock(&Mutex, MUTEX_NUM);

	/*
	 * Map mb memory to userspace
	 */
	mem_map_t mb_map;

	if (maps(memfd, 0x28000000, &mb_map, 0x8000000)) {
		printf("Mb_map mapping to userspace failed\r\n");
		exit(1);
	}

	sds_mmap((void *) 0x28000000, 0x8000000, (void *) mb_map.dev_base);

	/*
	 * Write firmware for the EdkDSP at mapped 0x28000000
	 * It is up to 64 firmware programs, each 16kbytes = 4096 uint32 values
	 */

	printf("ARMCPU0: Write firmware ... \n\r");

	FILE *fr0, *fr1, *fr2, *fr3;

	/*
     * open input files to read
     */
    printf("ARMCPU0: Open  input file f0.dec ... ");
    if ( (fr0 = fopen("./f0.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f1.dec ... ");
    if ( (fr1 = fopen("./f1.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

	printf("ARMCPU0: Open  input file f2.dec ... ");
    if ( (fr2 = fopen("./f2.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f3.dec ... ");
    if ( (fr3 = fopen("./f3.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    iiii = 0;
    value = 139266;
    while (fscanf(fr0,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
	}

    iiii = 0;
    value = 139266;
    while (fscanf(fr1,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
	}

	iiii = 0;
    value = 139266;
    while (fscanf(fr2,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
	}
    
	iiii = 0;
    value = 139266;
    while (fscanf(fr3,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
	}

	/*
	 * close input files
	 */
	printf("ARMCPU0: Close input file f0.dec ... ");
	if ( fclose(fr0) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f1.dec ... ");
	if ( fclose(fr1) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f2.dec ... ");
	if ( fclose(fr2) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f3.dec ... ");
	if ( fclose(fr3) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Write firmware Done. \n\r");

	/*
	 * Initial waiting loop for Microblaze mapped at 0x28100000
	 */

#if (1)
	printf("ARMCPU0: place 0x28100000 at start of MB0 vectors\n\r");
	Xil_Out32(mb_map.dev_base + 0x00100000, 0xb8000000);
#endif

	/*
	 * Remove Microblaze reset via system control of an emio pin
	 */

	start_Microblaze();

	printf("ARMCPU0: MB0 reset removed, ARM waiting ...\n\r");

	/*
	 * Wait for response from Microblaze via user value at mutex 0
	 */
	do {
		XMutex_GetUser(&Mutex, MUTEX_NUM, &val);
	} while (val == 0);

	printf("ARMCPU0: MB0 indicates - running ...\n\r");

	FILE *f;
	pix_t *inY1 = (pix_t *)sds_alloc_non_cacheable(HEIGHT*WIDTH);
	yuv_t *inCY1 = (yuv_t *)sds_alloc_non_cacheable(HEIGHT*WIDTH*2);
	pix_t *inY2 = (pix_t *)sds_alloc_non_cacheable(HEIGHT*WIDTH);
	yuv_t *inCY2 = (yuv_t *)sds_alloc_non_cacheable(HEIGHT*WIDTH*2);
	yuv_t *outCY = (yuv_t *)sds_alloc_non_cacheable(HEIGHT*WIDTH*2);
	printf("allocated buffers\n");

	f = fopen(FILEINAME,"rb");
	if (f == NULL) {
		printf("failed to open file %s\n", FILEINAME);
		return -1;
	}
	printf("opened file %s\n", FILEINAME);

	read_yuv_frame(inY1, WIDTH, WIDTH, HEIGHT, f);
	printf("read 1st %dx%d frame\n", WIDTH, HEIGHT);
	read_yuv_frame(inY2, WIDTH, WIDTH, HEIGHT, f);
	printf("read 2nd %dx%d frame\n", WIDTH, HEIGHT);

	fclose(f);
	printf("closed file %s\n", FILEINAME);

	convert_Y8toCY16(inY1, inCY1, HEIGHT*WIDTH);
	printf("converted 1st frame to 16bit\n");
	convert_Y8toCY16(inY2, inCY2, HEIGHT*WIDTH);
	printf("converted 2nd frame to 16bit\n");

	fpga_optflow(inCY1, inCY2, outCY, HEIGHT, WIDTH, WIDTH, 10.0);
	printf("computed optical flow\n");

	// write optical flow data image to disk
	write_yuv_file(outCY, WIDTH, WIDTH, HEIGHT, ONAME);

	sds_free(inY1);
	sds_free(inCY1);
	sds_free(inY2);
	sds_free(inCY2);
	sds_free(outCY);
	printf("freed buffers\n");

	printf("TEST PASSED\n");
	
	/*
	 *  unmap the memory before exiting
	 */
	if (munmap(mb_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	if (munmap(mutex_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	close(memfd);

	return 0;
}
