/*********** 
# Copyright (c) 2017, Xilinx, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT,INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY,OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
************/

/*******************************************************************************

    This is a vector addition example which showcases effectiveness of 
    reordering the loops helps in achieving better pipeline initiation 
    interval(II) and better performance.

*******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <cstdio>
#include <stdlib.h>
#include <xil_io.h>
#include <unistd.h>
#include "xparameters.h"
#include "xmutex.h"
#include <cstring>

#include "mmult.h"
#include "sds_utils.h"


#include "mem_map.h"

#define MUTEX_DEVICE_ID		XPAR_MUTEX_0_IF_0_DEVICE_ID

/*
 * The Mutex number on which the example is run
 */
#define MUTEX_NUM 		0

/*
 * Mutex instance
 */
XMutex Mutex;

void start_Microblaze() {
    // file handles
    int valuefd, exportfd, directionfd;

    // The GPIO has to be exported to be able to see it
    // in sysfs

    exportfd = open("/sys/class/gpio/export", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to export\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);

    // Update the direction of the GPIO to be an output

    directionfd = open("/sys/class/gpio/gpio960/direction", O_RDWR);
    if (directionfd < 0)
    {
        printf("Cannot open GPIO direction it\n");
        exit(1);
    }

    write(directionfd, "out", 4);
    close(directionfd);

    // Get the GPIO value ready to be toggled

    valuefd = open("/sys/class/gpio/gpio960/value", O_RDWR);
    if (valuefd < 0)
    {
        printf("Cannot open GPIO value\n");
        exit(1);
    }

    // reset Microblaze
    write(valuefd,"0", 2);

    printf("Reset for 1 sec. ...\n");
    sleep(1);
    printf("Reset for 1 sec. done\n");

    // release MB reset
    write(valuefd,"1", 2);

    close(valuefd);

    //release exported gpio pin
    exportfd = open("/sys/class/gpio/unexport", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to unexport\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);
    return;
}



// Software implementation of Matrix Multiplication
// The inputs are of the size (DATA_SIZE x DATA_SIZE)
void m_softwareGold(
                    int *in1,   //Input Matrix 1
                    int *in2,   //Input Matrix 2
                    int *out    //Output Matrix
                   )
{
    //Perform Matrix multiply Out = In1 x In2
    for(int i = 0; i < DATA_SIZE; i++) {
        for(int j = 0; j < DATA_SIZE; j++) {
            for(int k = 0; k < DATA_SIZE; k++) {
                out[i * DATA_SIZE + j] += in1[i * DATA_SIZE + k] * in2[k * 
                                          DATA_SIZE + j];
            }
        }
    }
}

int main(int argc, char** argv)
{


	/*
	 * Init mutex and start microblaze
	 */
	XMutex_Config *ConfigPtr;
	XStatus Status;
	u32 TimeoutCount = 0;
	u32 val;

    u32 iiii=0;
    u32 value;
	
	int memfd;
	memfd = open("/dev/mem", O_RDWR | O_SYNC);
	if (memfd == -1) {
		printf("Can't open /dev/mem.\n");
		exit(0);
	}
	printf("/dev/mem opened.\n");

	/*
	 * Lookup configuration data in the device configuration table.
	 * Use this configuration info down below when initializing this
	 * driver instance.
	 */

	ConfigPtr = XMutex_LookupConfig(MUTEX_DEVICE_ID);
	if (ConfigPtr == (XMutex_Config *) NULL) {
		return XST_FAILURE;
	}

	/*
	 * Map Mutex memory to userspace
	 */
	mem_map_t mutex_map;

	if (map(memfd, (off_t) ConfigPtr->BaseAddress, &mutex_map)) {
		printf("Mutex mapping to userspace failed\r\n");
		exit(1);
	}

	ConfigPtr->BaseAddress = mutex_map.dev_base;

	/*
	 * Perform the rest of the initialization.
	 */
	Status = XMutex_CfgInitialize(&Mutex, ConfigPtr, ConfigPtr->BaseAddress);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XMutex_SetUser(&Mutex, MUTEX_NUM, 0);
	XMutex_Lock(&Mutex, MUTEX_NUM);

	/*
	 * Map mb memory to userspace
	 */
	mem_map_t mb_map;

	if (maps(memfd, 0x28000000, &mb_map, 0x8000000)) {
		printf("Mb_map mapping to userspace failed\r\n");
		exit(1);
	}

	sds_mmap((void *) 0x28000000, 0x8000000, (void *) mb_map.dev_base);

	/*
	 * Write firmware for the EdkDSP at mapped 0x28000000
	 * It is up to 64 firmware programs, each 16kbytes = 4096 uint32 values
	 */

	printf("ARMCPU0: Write firmware ... \n\r");

	FILE *fr0, *fr1, *fr2, *fr3;

	/*
     * open input files to read
     */
    printf("ARMCPU0: Open  input file f0.dec ... ");
    if ( (fr0 = fopen("./f0.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f1.dec ... ");
    if ( (fr1 = fopen("./f1.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

	printf("ARMCPU0: Open  input file f2.dec ... ");
    if ( (fr2 = fopen("./f2.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f3.dec ... ");
    if ( (fr3 = fopen("./f3.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    iiii = 0;
    value = 139266;
    while (fscanf(fr0,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
	}

    iiii = 0;
    value = 139266;
    while (fscanf(fr1,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
	}

	iiii = 0;
    value = 139266;
    while (fscanf(fr2,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
	}
    
	iiii = 0;
    value = 139266;
    while (fscanf(fr3,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
	}

	/*
	 * close input files
	 */
	printf("ARMCPU0: Close input file f0.dec ... ");
	if ( fclose(fr0) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f1.dec ... ");
	if ( fclose(fr1) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f2.dec ... ");
	if ( fclose(fr2) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f3.dec ... ");
	if ( fclose(fr3) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Write firmware Done. \n\r");

	/*
	 * Initial waiting loop for Microblaze mapped at 0x28100000
	 */

#if (1)
	printf("ARMCPU0: place 0x28100000 at start of MB0 vectors\n\r");
	Xil_Out32(mb_map.dev_base + 0x00100000, 0xb8000000);
#endif

	/*
	 * Remove Microblaze reset via system control of an emio pin
	 */

	start_Microblaze();

	printf("ARMCPU0: MB0 reset removed, ARM waiting ...\n\r");

	/*
	 * Wait for response from Microblaze via user value at mutex 0
	 */
	do {
		XMutex_GetUser(&Mutex, MUTEX_NUM, &val);
	} while (val == 0);

	printf("ARMCPU0: MB0 indicates - running ...\n\r");


    if (DATA_SIZE > MAX_SIZE) {
        std::cout << "Size is bigger than internal buffer size, please use a size"; 
        std::cout << " smaller than " << MAX_SIZE << "!" << std::endl;
        return 1;
    }

    // Size of the matrix
    size_t matrix_size_bytes = sizeof(int) * DATA_SIZE * DATA_SIZE;

    // Allocate buffers using sds_alloc
    int *source_in1         = (int *) sds_alloc(matrix_size_bytes);
    int *source_in2         = (int *) sds_alloc(matrix_size_bytes);
    int *source_hw_results  = (int *) sds_alloc(matrix_size_bytes);

    // Allocate software output buffer
    int *source_sw_results  = (int *) malloc(matrix_size_bytes);

    // Check for failed memory allocation
    if((source_in1 == NULL) || (source_in2 == NULL) || (source_hw_results == NULL) || (source_sw_results == NULL)){
        std::cout << "TEST FAILED : Failed to allocate memory" << std::endl;
        return -1;
    }

    // Create the test data 
    for(int i = 0 ; i < DATA_SIZE * DATA_SIZE ; i++){
        source_in1[i] = i;
        source_in2[i] = i * i;
        source_sw_results[i] = 0;
        source_hw_results[i] = 0;
    }
    int size = DATA_SIZE;

    sds_utils::perf_counter hw_ctr, sw_ctr;

    hw_ctr.start();
    // Launch Hardware Solution 
    mmult_accel(source_in1, source_in2, source_hw_results, size);
    hw_ctr.stop();

    sw_ctr.start();
    // Launch Software Solution
    m_softwareGold(source_in1, source_in2, source_sw_results);
    sw_ctr.stop();

    uint64_t sw_cycles = sw_ctr.avg_cpu_cycles();
    uint64_t hw_cycles = hw_ctr.avg_cpu_cycles();
    double speedup = (double) sw_cycles / (double) hw_cycles;

    std::cout << "Number of CPU cycles running application in software: "
                << sw_cycles << std::endl;
    std::cout << "Number of CPU cycles running application in hardware: "
                << hw_cycles << std::endl;
    std::cout << "Speed up: " << speedup << std::endl;
    std::cout << "Note: Speed up is meaningful for real hardware execution only, not for emulation." << std::endl;

    // Compare the results 
    bool match = true;
    for (int i = 0 ; i < DATA_SIZE * DATA_SIZE ; i++){
        if (source_hw_results[i] != source_sw_results[i]){
            std::cout << "Error: Result mismatch" << std::endl;
            std::cout << "i = " << i << " CPU result = " << source_sw_results[i]
                << " Hardware result = " << source_hw_results[i] << std::endl;
            match = false;
            break;
        }
    }

    // Release memory
    sds_free(source_in1);
    sds_free(source_in2);
    sds_free(source_hw_results);
    free(source_sw_results);

    if (!match){
        std::cout << "TEST FAILED." << std::endl;
        return 1;
    }
    std::cout << "TEST PASSED." << std::endl;


	/*
	 *  unmap the memory before exiting
	 */
	if (munmap(mb_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	if (munmap(mutex_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	close(memfd);

    return 0;
}
